#include "includes.h"
#include "support.h"
#include "packet.h"
#include "dhcp.h"

//#define __RECORD__

#ifdef  DEBUG
#define DHCP_DEBUG_OUT(s) Handler(DHCP_DEBUG,s,0);
#else
#define DHCP_DEBUG_OUT(s)
#endif

void    DHCPLeaseData::Reset()
{
        memset(this,0,sizeof(DHCPLeaseData));
}

void    DHCPSessionData::Reset()
{
        memset(this,0,sizeof(DHCPSessionData));
        static unsigned char reqopts[] = {51,28,15,12,6,3,1,0};
        memcpy(RequestOptions,reqopts,sizeof(reqopts));
}


DHCP::DHCPSocket::~DHCPSocket()
{
#ifdef  __WIN32__
        closesocket(s);
#else
        close(s);
#endif
}

DHCP::DHCPUdpSocket::DHCPUdpSocket(unsigned short port,unsigned char *ip)
{
        int n;
        sockaddr sa = {AF_INET,""};
        memset(sa.sa_data,0,sizeof(sa.sa_data));

        s = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
        n = 1;
        setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *) &n,sizeof(n));
        n = 1;
        setsockopt(s,SOL_SOCKET,SO_BROADCAST,(char *) &n,sizeof(n));

        ((unsigned short *) sa.sa_data) [0] = htons(port);
        memcpy(&sa.sa_data[2],ip,4);
        bind(s,&sa,sizeof(sa));
}

int     DHCP::DHCPUdpSocket::Send(void *data,
                                  int len,
                                  unsigned short port,
                                  const unsigned char *dstip)
{
        sockaddr sa;
        memset(&sa,0,sizeof(sa));
        sa.sa_family = AF_INET;
        ((unsigned short *) sa.sa_data) [0] = htons(port);
        memcpy(&sa.sa_data[2],dstip,4);

        return sendto(s,(char *) data,len,0,&sa,sizeof(sa));
}

int     DHCP::DHCPUdpSocket::Recv(void *data,
                                  int len,
                                  unsigned short &port,
                                  unsigned char *srcip)
{
        sockaddr sa;
#ifdef  __WIN32__
        int n = sizeof(sa);
#else
        socklen_t n = sizeof(sa);
#endif
        int c = recvfrom(s,(char *) data,len,0,&sa,&n);

        memcpy(srcip,&sa.sa_data[2],4);
        port = ntohs(((unsigned short *) sa.sa_data) [0]);
        return c;
}

int     DHCP::DHCPUdpSocket::CheckDup(const unsigned char *ip,const unsigned char *)
{
        ICMPPing p;
        return p.Ping(ip,1000,1);
}

void    DHCP::DHCPUdpSocket::Unarp()
{
}

#ifndef __WIN32__

DHCP::DHCPRawSocket::DHCPRawSocket(char *iface,
                                   unsigned char *hw,
                                   int len,
                                   unsigned short port,
                                   unsigned int cp,
                                   unsigned char *ip,
                                   unsigned char *nm,
                                   unsigned char *gw)
{
        Interface = iface;
        HWAddress = hw;
        HALength = len;
        Port = port;

        s = -1;
        if(s == -1) s = socket(AF_PACKET,SOCK_PACKET,htons(ETH_P_ALL));
        if(s == -1) s = socket(AF_INET,SOCK_PACKET,htons(ETH_P_ALL));

        int n;
        n = 1;
        setsockopt(s,SOL_SOCKET,SO_BROADCAST,(char *) &n,sizeof(n));

        ConformParms = cp;

        IP = ip;
        Netmask = nm;
        Gateway = gw;
}

int     DHCP::DHCPRawSocket::TestPacket(unsigned char *buf,int len,PacketInfo &pi)
{
        int pos;
        unsigned char *b = buf;
        unsigned int blen = len;

        pos = DecodeEther(b,blen);
        if(pos == 0) return 0;
        ETHERHdr &eh = ((ETHERHdr *) b) [0];
        if(ntohs(eh.Type) != __ETHERTYPE_IP) return 0;
        b += pos;
        blen -= pos;

        pos = DecodeIp(b,blen);
        if(pos == 0) return 0;
        IPHdr &ih = ((IPHdr *) b) [0];
        b += pos;
        blen -= pos;

        pos = DecodeUdp(b,blen,ih);
        if(pos == 0) return 0;
        UDPHdr &uh = ((UDPHdr *) b) [0];
        if(ntohs(uh.DestPort) != Port) return 0;
        b += pos;
        blen -= pos;

        pi.eh = &eh;
        pi.ih = &ih;
        pi.uh = &uh;
        pi.data = b;

        return blen;
}

const unsigned char * DHCP::DHCPRawSocket::GetTarget(const unsigned char *ip)
{
        if(memcmp(IP,"\x0\x0\x0\x0",4) == 0) return 0;

        int c = 0;
        while(c < 4) {
                if((IP[c] & Netmask[c]) != (ip[c] & Netmask[c])) return Gateway;
                c++;
        };
        return ip;
}

int     DHCP::DHCPRawSocket::Send(void *p,
                                  int len,
                                  unsigned short port,
                                  const unsigned char *destip)
{
        unsigned char buf[1500];
        unsigned char *ptr = buf;
        int pos;
        int blen = 1500,slen;
        unsigned char destHW[6];
        sockaddr sa;

        if(ConformParms & DHCP_C_SEND_BCAST_PACKETS ||
           memcmp(destip,"\xFF\xFF\xFF\xFF",4) == 0) {
                memset(destHW,0xFF,6);
        }
        else {
                ARP a;
                unsigned char *hw;
                const unsigned char *tip;
                tip = GetTarget(destip);
//#ifdef  __RECORD__
//                printf("IP: %d.%d.%d.%d\n",tip[0],tip[1],tip[2],tip[3]);
//#endif
                if(tip != 0 &&
                   (hw = a.GetHW(s,Interface,tip,IP,HWAddress,5)) != 0) {
                        memcpy(destHW,hw,6);
                }
                else memset(destHW,0xFF,6);
        };

        pos = EncodeIpEther(ptr,blen,destHW,HWAddress);
        if(pos == 0) return 0;
        ptr += pos;
        blen -= pos;
        pos = EncodeUdpIp(ptr,blen,p,len,port,destip,Port,IP);
        if(pos == 0) return 0;
        ptr += pos;
        blen = ptr - buf;

        memset(&sa,0,sizeof(sa));
        strcpy(sa.sa_data,Interface);

#ifdef  __RECORD__
        FILE *f = fopen("send.out","a+");
        fwrite(&blen,4,1,f);
        fwrite(buf,blen,1,f);
        fclose(f);
#endif

        slen = sendto(s,buf,blen,0,&sa,sizeof(sa));
        if(slen != blen) {
                return 0;
        };

        return len;
}

int     DHCP::DHCPRawSocket::Recv(void *data,
                                  int len,
                                  unsigned short &port,
                                  unsigned char *srcip)
{
        struct PreCheck {
                ETHERHdr eh;
                IPHdr ih;
        };

        unsigned char buf[1500];
        PreCheck *pc = (PreCheck *) buf;
        sockaddr sa;
        socklen_t n;
        int blen;
        PacketInfo pi;

        n = sizeof(sa);
        blen = recvfrom(s,buf,1500,0,&sa,&n);
        if(blen <= 0) return 0;
        if(pc[0].ih.Protocol != IPPROTO_UDP) return 0;
        if(strcmp(sa.sa_data,Interface) != 0) return 0;

#ifdef  __RECORD__
        FILE *f = fopen("recv.out","a+");
        fwrite(&blen,4,1,f);
        fwrite(buf,blen,1,f);
        fclose(f);
#endif

        blen = TestPacket(buf,blen,pi);
        if(blen > len) blen = len;
        memcpy(data,pi.data,blen);
        return blen;
}

int     DHCP::DHCPRawSocket::CheckDup(const unsigned char *ip,const unsigned char *sip)
{
        ARP a;
        return a.GetHW(s,Interface,ip,sip,HWAddress,1) != NULL;
}

void    DHCP::DHCPRawSocket::Unarp()
{
        SendUnarp(s,Interface,HWAddress,IP);
}

#endif


IDF     DHCP::RandomID(1);

unsigned char * DHCP::ConvertIP(char *s)
{
        static unsigned char IP[4];
        ((unsigned long *) IP) [0] = inet_addr(s);
        return IP;
}

void    DHCP::Handler(int,char *,char *)
{
}

void    DHCP::SetAuthName(char *s)
{
        strcpy((char *) AuthName,s);
}

void    DHCP::SetClassID(char *s)
{
        strcpy((char *) ClassID,s);
}

void    DHCP::StoreList(unsigned char *src,unsigned char *dst,int ilen,int n,int max)
{
        int     c;
        c = 0;
        while(c < n && c < max) {
                memcpy(&dst[c*ilen],&src[c*ilen],ilen);
                c++;
        };
        while(c < max) {
                memset(&dst[c*ilen],0,ilen);
                c++;
        };
}

void    DHCP::StoreTime()
{
        unsigned long *t;
        t = (unsigned long *) GetOption(51);
        if(t) {
                RenewTime = ntohl(t[0]);
        }
        else RenewTime = 0;
}

void    DHCP::StoreOptions()
{
        int c,n;
        unsigned char *s;

        s = GetOption(1);
        if(s) memcpy(Netmask,s,4);

        s = GetOption(3);
        if(s) StoreList(s,Routers[0],4,s[-1]/4,4);

        s = GetOption(6);
        if(s) StoreList(s,DomainNameServers[0],4,s[-1]/4,4);

        s = GetOption(12);
        if(s) {
                n = s[-1];
                if(n >= (int) sizeof(HostName)) n = sizeof(HostName) - 1;
                memcpy(HostName,s,n);
                HostName[n] = 0;
        };
        s = GetOption(15);
        if(s) {
                n = s[-1];
                if(n >= (int) sizeof(DomainName)) n = sizeof(DomainName) - 1;
                memcpy(DomainName,s,n);
                DomainName[n] = 0;
        };
        s = GetOption(28);
        if(s) memcpy(Broadcast,s,4);
        else {
                c = 0;
                while(c < 4) {
                        Broadcast[c] = LocalIP[c];
                        Broadcast[c] &= Netmask[c];
                        Broadcast[c] ^= Netmask[c] ^ (unsigned char) 0xFF;
                        c++;
                };
        };
}

void    DHCP::StoreBasic(unsigned char *ip,unsigned char *sid)
{
        if(ip) memcpy(LocalIP,ip,4);
        if(sid) memcpy(ServerIP,sid,4);
}

unsigned char *  DHCP::GetOption(int opval)
{
        unsigned char *ptr = Packet.options;

        while(ptr[0] != 255 || opval == 255) {
                if(ptr[0] == 0) {
                        ptr++;
                        continue;
                };
                if(ptr[0] == 255) return ptr;
                if(ptr[0] == opval) return ptr + 2;

                ptr++;
                ptr += ptr[0];
                ptr++;
        };

        return 0;
}

int     DHCP::OpenChannel()
{
        if(inet_addr(Interface) != INADDR_NONE) {
                union {
                        unsigned char   ip[4];
                        unsigned long   l;
                };
                l = inet_addr(Interface);

                Socket = new DHCPUdpSocket(ClientPort,ip);
        }
        else {
#ifndef __WIN32__
                Socket = new DHCPRawSocket(Interface,
                                           HardwareAddress,
                                           HALength,
                                           ClientPort,
                                           ConformParms,
                                           LocalIP,
                                           Netmask,
                                           Routers[0]);
#endif
        };

        if(Socket == 0) return 0;

        if(Socket[0].s == -1) {
                delete Socket;
                Socket = 0;
                return 0;
        };

        return 1;
}

void    DHCP::CloseChannel()
{
        if(Socket) {
                delete Socket;
                Socket = 0;
        };
}

int     DHCP::Send(unsigned char *ip)
{
        if(Socket == 0) return -1;

        return Socket[0].Send(&Packet,sizeof(Packet),DHCP_SERVER_PORT,ip);

//        fwrite(Raw,c,1,stdout);
}

unsigned long   DHCP::GetID()
{
        return RandomID.GetID();
}

int     DHCP::WaitResponse(unsigned long id,unsigned long msecs)
{
        if(Socket == 0) return 0;
        
        fd_set fd;
        timeval tv;
        tv.tv_sec = msecs / 1000;
        tv.tv_usec = (msecs % 1000) * 1000;
        unsigned char sip[4];

        FD_ZERO(&fd);
        FD_SET(Socket[0].s,&fd);

        if(select(Socket[0].s+1,&fd,0,0,&tv) == 0) return 0;
        if(FD_ISSET(Socket[0].s,&fd) == 0) return 0;
        if(Socket[0].Recv(&Packet,sizeof(Packet),ClientPort,sip) <= 0) return 0;

        if(Packet.op != 2) return 0;
        if(memcmp(Packet.chaddr,HardwareAddress,HALength) != 0) return 0;
        if(Packet.xid != id) return 0;

//        if(Packet.op == 2 && Packet.xid == id &&
//           memcmp(Packet.chaddr,HardwareAddress,HALength) == 0) return 1;

        return 1;
}

void    DHCP::AddOption(unsigned char opt,int len,const unsigned char *o)
{
        unsigned char *ptr = GetOption(255);

        ptr[0] = opt;
        ptr++;
        ptr[0] = len;
        ptr++;
        if(len) memcpy(ptr,o,len);
        ptr += len;
        if(len & 1) {
                ptr[0] = 0;
                ptr++;
        };
        ptr[0] = 255;
}

void    DHCP::AddOption(unsigned char opt,int len,int ch,...)
{
        unsigned char *ptr = GetOption(255);
        int *o = &ch;
        int c = 0;

        ptr[0] = opt;
        ptr++;
        ptr[0] = len;
        ptr++;
        while(c < len) {
                ptr[0] = o[0];
                o++;
                ptr++;
                c++;
        };
        if(len & 1) {
                ptr[0] = 0;
                ptr++;
        };
        ptr[0] = 255;
}

void    DHCP::AddOption(unsigned char opt,int len,long i,...)
{
        AddOption(opt,len,(char *) &i);
}
void    DHCP::AddOption(unsigned char opt,int len,unsigned long i,...)
{
        AddOption(opt,len,(char *) &i);
}

void    DHCP::AddOption(unsigned char opt,int len,const char *p)
{
        AddOption(opt,len,(unsigned char *) p);
}

unsigned long   DHCP::PrepRequest()
{
        memset(&Packet,0,sizeof(Packet));
        Packet.op = 1;
        Packet.htype = HAType;
        Packet.hlen = HALength;
        Packet.hops = 0;
        Packet.xid = GetID();
        Packet.secs = 0;
        if(ConformParms & DHCP_C_REQ_BCAST_REPLY_NEVER) Packet.flags = 0;
        else Packet.flags = htons(0x8000);
//        Packet.flags = htons(0x8001);
//        flags = htons(1);
        memset(Packet.ciaddr,0,4);
        memcpy(Packet.chaddr,HardwareAddress,16);
        Packet.sname[0] = 0;
        Packet.file[0] = 0;
        memcpy(Packet.Cookie,"\x63\x82\x53\x63",4);
        Packet.options[0] = 255;

        if(memcmp(PrimeIP,"\x7F\x0\x0\x1",4) == 0) {
#ifndef __WIN32__
                if(inet_addr(Interface) == INADDR_NONE) {
                        struct ifreq ifr;
                        int s;
                        s = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
                        strncpy(ifr.ifr_name,Interface,sizeof(ifr.ifr_name));
                        ifr.ifr_addr.sa_family = AF_INET;
                        if(ioctl(s,SIOCGIFADDR,&ifr) == 0) {
                                memcpy(Packet.giaddr,&ifr.ifr_addr.sa_data[2],4);
                        }
                        else memset(Packet.giaddr,0,4);
                        close(s);
                }
                else memset(Packet.giaddr,0,4);
#endif
        }
        else memcpy(Packet.giaddr,PrimeIP,4);

        return Packet.xid;
}

void    DHCP::AddRequestOptions(unsigned char *ops)
{
        unsigned char *src = ops;
        while(*src) {
                unsigned char *dst = RequestOptions;
                while(*dst && *src != *dst) dst++;
                if(!*dst) {
                        *dst = *src;
                        dst++;
                        *dst = 0;
                };
                src++;
        };
}

void    DHCP::AddOptions(unsigned char *list)
{
        unsigned char *ptr = list;

        while(ptr[0] != 255) {
                AddOption(ptr[0],ptr[1],&ptr[2]);
                ptr++;
                ptr += ptr[0];
                ptr++;
        };
}

void    DHCP::SetOptions(unsigned char *ops)
{
        memcpy(Options,ops,128);
}

void    DHCP::ConfigureInterface(char *iface,int type,int len,const unsigned char *cid,int clen)
{
        if(HAType) {
                HAType = type;
                HALength = len;
        }
        else {
                HAType = 1;             //defaults
                HALength = 6;
        };

        if(iface != 0 && inet_addr(iface) == INADDR_NONE) {
#ifndef  __WIN32__
//we can do a whole bunch of stuff here if in linux
//detect the HWaddr
                strncpy(Interface,iface,sizeof(Interface));

                struct ifreq ifr;
                int s;

                s = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
                strncpy(ifr.ifr_name,iface,sizeof(ifr.ifr_name));
                if(ioctl(s,SIOCGIFHWADDR,&ifr) == 0) {
                        memcpy(HardwareAddress,ifr.ifr_hwaddr.sa_data,HALength);
                };
                close(s);
#endif
        }
        else {
                if(iface) strncpy(Interface,iface,sizeof(Interface));
                else strcpy(Interface,"0.0.0.0");
                memcpy(HardwareAddress,cid,HALength);
        };

        if(cid) {
                memcpy(ClientID,cid,clen); //we can override the network address
                CIDLength = clen;
        }
        else {
                ClientID[0] = (unsigned char) HAType;
                memcpy(&ClientID[1],HardwareAddress,HALength);
                CIDLength = HALength + 1;
        };
}

int     DHCP::Request(const unsigned char *cip,
                      const unsigned char *ip,
                      const unsigned char *server,
                      int bflag,
                      int replytype)
{
        unsigned long i;
        unsigned long t,w;
        unsigned char *opsend;
//        unsigned char *sid;

        i = PrepRequest();
        AddOption(53,1,DHCP_REQUEST);
        if(ClassID[0]) AddOption(60,strlen(ClassID)+1,ClassID);
        AddOption(61,CIDLength,ClientID);
        t = htonl(PrefTime);
        AddOption(51,4,t);
//        AddOption(55,7,51,28,15,12,6,3,1);
        opsend = RequestOptions;
        while(*opsend) opsend++;
        AddOption(55,opsend - RequestOptions,RequestOptions);
//        printf("%s: %s",Interface,AuthName);
        if(AuthName[0]) AddOption(12,strlen(AuthName) + 1,AuthName);

        if(server) AddOption(54,4,server);
        if(ip) AddOption(50,4,ip);
        if(cip) memcpy(Packet.ciaddr,cip,4);
        if(replytype == 0) Packet.flags = 0;
        if(replytype == -1) Packet.flags = htons(0x8000);

        AddOptions(Options);

        if(OpenChannel() == 0) return -1;
        if(bflag) Send(BroadcastAddress);
        else Send(ServerIP);

        t = GetTicks();
        while((w = (GetTicks() - t)) < RequestTimeout) {
                w = RequestTimeout - w;
                if(WaitResponse(i,w) != 0) {
                        CloseChannel();
                        unsigned char *s = GetOption(53);
                        if(s != 0 && s[0] == DHCP_ACK) {
                                return 1;
                        };
                        return 0;
                };
        };
        CloseChannel();
        return -1;
}


int     DHCP::Inform()
{
        int setip = 0;

        if(memcmp(LocalIP,"\x0\x0\x0\x0",4) == 0) {
                setip = 1;
                if(inet_addr(Interface) == INADDR_NONE) {
#ifndef __WIN32__
                        struct ifreq ifr;
                        int s;

                        s = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
                        strncpy(ifr.ifr_name,Interface,sizeof(ifr.ifr_name));
                        if(ioctl(s,SIOCGIFADDR,&ifr) == 0) {
                                memcpy(LocalIP,&ifr.ifr_addr.sa_data[2],4);
                        };
                        close(s);
#endif
                }
                else if(strcmp(Interface,"0.0.0.0") != 0) {
                        ((unsigned long *) LocalIP) [0] = inet_addr(Interface);
                }
                else if(memcmp(PrimeIP,"\x0\x0\x0\x0",4) != 0) {
                        memcpy(LocalIP,PrimeIP,4);
                }
                else return -2;
        };

        Handler(DHCP_INFORM,"\xFF\xFF\xFF\xFF",(char *) LocalIP);

        unsigned long i;
        unsigned long t,w;

        i = PrepRequest();
        AddOption(53,1,DHCP_INFORM);
        if(ClassID[0]) AddOption(60,strlen(ClassID)+1,ClassID);
        AddOption(61,CIDLength,ClientID);
        AddOption(55,4,15,12,6,3);
//        printf("%s: %s",Interface,AuthName);
        if(AuthName[0]) AddOption(12,strlen(AuthName) + 1,AuthName);
        memcpy(Packet.ciaddr,LocalIP,4);
        Packet.flags = 0;

        AddOptions(Options);

        if(OpenChannel() == 0) return -1;
        Send(BroadcastAddress);

        t = GetTicks();
        while((w = (GetTicks() - t)) < RequestTimeout) {
                w = RequestTimeout - w;
                if(WaitResponse(i,w) != 0) {
                        CloseChannel();
                        unsigned char *res = GetOption(53);

                        if(res && res[0] == DHCP_ACK) {
                                StoreOptions();
                                Handler(DHCP_ACK,0,0);
                                return 1;
                        };
                        if(setip) memset(LocalIP,0,4);
                        Handler(DHCP_NACK,(char *) GetOption(56),0);
                        return 0;
                };
        };
        CloseChannel();
        if(setip) memset(LocalIP,0,4);
        return -1;
}

int     DHCP::IsSelectee(int pos,const IPName &ip)
{
        if(pos >= nOffers) return -1;
        if(memcmp(ip.Address,Offered[pos],4) == 0) return 1;
        return 0;
}

int     DHCP::Select(int pos)
{
        if(pos >= nOffers) return 0;
        int ret;

        if(ConformParms & DHCP_C_CHECK_IP_USE &&
           OpenChannel() != 0) {
                const unsigned char *chkip = Offered[pos];
                ret = Socket[0].CheckDup(chkip,chkip);
                CloseChannel();
                if(ret != 0) {
                        Decline(chkip);
                        return 0;
                };
        };

        Handler(DHCP_REQUEST_SELECT,(char *) Servers[pos],(char *) Offered[pos]);

        ret = Request(0,Offered[pos],Servers[pos],1,1);
        if(ret > 0) {
                StoreBasic(Packet.yiaddr,GetOption(54));
                StoreTime();
                StoreOptions();
                Handler(DHCP_ACK,0,0);
        };
        if(ret == 0) Handler(DHCP_NACK,(char *) GetOption(56),(char *) Offered[pos]);
        return ret;
}

int     DHCP::Select(int pos,const IPName &ip)
{
        if(pos >= nOffers) return 0;
        int ret;

        if(ConformParms & DHCP_C_CHECK_IP_USE &&
           OpenChannel() != 0) {
                const unsigned char *chkip = ip.Address;
                ret = Socket[0].CheckDup(chkip,chkip);
                CloseChannel();
                if(ret != 0) return 0;
        };

        Handler(DHCP_REQUEST_SELECT,(char *) Servers[pos],(char *) ip.Address);

        ret = Request(0,ip.Address,Servers[pos],1,1);
        if(ret > 0) {
                StoreBasic(Packet.yiaddr,GetOption(54));
                StoreTime();
                StoreOptions();
                Handler(DHCP_ACK,0,0);
        };
        if(ret == 0) Handler(DHCP_NACK,(char *) GetOption(56),(char *) ip.Address);
        return ret;
}

int     DHCP::Select(const IPName &server,const IPName &ip)
{
        int ret;

        if(ConformParms & DHCP_C_CHECK_IP_USE &&
           OpenChannel() != 0) {
                const unsigned char *chkip = ip.Address;
                ret = Socket[0].CheckDup(chkip,chkip);
                CloseChannel();
                if(ret != 0) return 0;
        };

        Handler(DHCP_REQUEST_SELECT,(char *) server.Address,(char *) ip.Address);

        ret = Request(0,ip.Address,server.Address,1,1);
        if(ret > 0) {
                StoreBasic(Packet.yiaddr,GetOption(54));
                StoreTime();
                StoreOptions();
                Handler(DHCP_ACK,0,0);
        };
        if(ret == 0) Handler(DHCP_NACK,(char *) GetOption(56),(char *) ip.Address);
        return ret;
}

int     DHCP::Reselect(const IPName &ip)
{
        int ret;

        if(ConformParms & DHCP_C_CHECK_IP_USE &&
           OpenChannel() != 0) {
                const unsigned char *chkip = ip.Address;
                ret = Socket[0].CheckDup(chkip,chkip);
                CloseChannel();
                if(ret != 0) return 0;
        };

        Handler(DHCP_REQUEST_RESELECT,"\xFF\xFF\xFF\xFF",(char *) ip.Address);

        ret = Request(0,ip.Address,0,1,1);
        if(ret > 0) {
                StoreBasic(Packet.yiaddr,GetOption(54));
                StoreTime();
                StoreOptions();
                Handler(DHCP_ACK,0,0);
        };
        if(ret == 0) Handler(DHCP_NACK,(char *) GetOption(56),(char *) ip.Address);
        return ret;
}

int     DHCP::Renew()
{
        int ret;

        Handler(DHCP_REQUEST_RENEW,0,0);

        ret = Request(LocalIP,0,0,0,ConformParms & DHCP_C_REQ_BCAST_REPLY_ALWAYS);
        if(ret > 0) {
                StoreBasic(Packet.yiaddr,GetOption(54));
                StoreTime();
                StoreOptions();
                Handler(DHCP_RENEW,0,0);
        };
        if(ret == 0) Handler(DHCP_NACK,(char *) GetOption(56),0);
        return ret;
}

int     DHCP::Rebind()
{
        int ret;

        Handler(DHCP_REQUEST_REBIND,0,0);

        ret = Request(LocalIP,0,0,1,ConformParms & DHCP_C_REQ_BCAST_REPLY_ALWAYS);
        if(ret > 0) {
                StoreBasic(Packet.yiaddr,GetOption(54));
                StoreTime();
                StoreOptions();
                Handler(DHCP_RENEW,0,0);
        };
        if(ret == 0) Handler(DHCP_NACK,(char *) GetOption(56),0);
        return ret;
}

void    DHCP::Release()
{
        if(memcmp(LocalIP,"\x0\x0\x0\x0",4) == 0) return;

        PrepRequest();
        AddOption(53,1,DHCP_RELEASE);
        AddOption(61,CIDLength,ClientID);
        memcpy(Packet.ciaddr,LocalIP,4);
        if(ClassID[0]) AddOption(60,strlen(ClassID)+1,ClassID);
//        printf("%s: %s",Interface,AuthName);
        if(AuthName[0]) AddOption(12,strlen(AuthName) + 1,AuthName);
        Packet.flags = 0;

        AddOptions(Options);

        Handler(DHCP_RELEASE,0,0);

        if(OpenChannel() == 0) return;
        Send(ServerIP);
        Socket[0].Unarp();
        CloseChannel();

        DHCPLeaseData::Reset();
}

void    DHCP::Decline(const unsigned char *ip)
{
        PrepRequest();
        AddOption(53,1,DHCP_DECLINE);
        AddOption(61,CIDLength,ClientID);
        AddOption(50,4,ip);
        if(ClassID[0]) AddOption(60,strlen(ClassID)+1,ClassID);
//        printf("%s: %s",Interface,AuthName);
        if(AuthName[0]) AddOption(12,strlen(AuthName) + 1,AuthName);

        AddOptions(Options);

        Send(BroadcastAddress);
        Handler(DHCP_DECLINE,(char *) ip,0);
}

void    DHCP::SetPrimeIP(const IPName &ip)
{
        if(ip != 0) {
                memcpy(PrimeIP,ip.Address,4);
                ClientPort = DHCP_SERVER_PORT;
        }
        else {
                memset(PrimeIP,0,4);
                ClientPort = DHCP_CLIENT_PORT;
        };
}

int     DHCP::Discover(const IPName &previp)
{
        unsigned char   (&servers)[4][4] = Servers;
        unsigned char   (&ips)[4][4] = Offered;

        unsigned long i;
        unsigned long t,w;
        unsigned char *sid;
        int n;
//        int c,d,ret;

        if(previp != 0) Handler(DHCP_DISCOVER,(char *) previp.Address,0);
        else Handler(DHCP_DISCOVER,0,0);

        i = PrepRequest();
        AddOption(53,1,DHCP_DISCOVER);
        t = htonl(PrefTime);
        AddOption(51,4,t);
        if(ClassID[0]) AddOption(60,strlen(ClassID)+1,ClassID);
        AddOption(61,CIDLength,ClientID);
        AddOption(55,1,54);
//        printf("%s: %s",Interface,AuthName);
        if(AuthName[0]) AddOption(12,strlen(AuthName) + 1,AuthName);
        if(previp != 0) AddOption(50,4,previp.Address);

        AddOptions(Options);

        if(OpenChannel() == 0) return -1;
        Send(BroadcastAddress);
        t = GetTicks();
        n = 0;
        while((w = (GetTicks() - t)) < DiscoverTimeout && n < 4) {
                w = DiscoverTimeout - w;
                unsigned char *res;
                if(WaitResponse(i,w) != 0 &&
                   (res = GetOption(53)) != 0 &&
                   res[0] == DHCP_OFFER &&
                   (sid = GetOption(54)) != 0) {
                        memcpy(servers[n],sid,4);
                        memcpy(ips[n],Packet.yiaddr,4);
                        Handler(DHCP_OFFER,(char *) sid,(char *) Packet.yiaddr);
                        n++;

                        if(ConformParms & DHCP_C_NOCOLLECT_OFFERS) break;
                        if(previp != 0) break;
                };
        };
        CloseChannel();

/*        c = 0;  //try and get the previous IP from some server
        while(c < n && previp != 0) {
                if(Request(0,previp.Address,servers[c],1,1) > 0) {
                        StoreBasic(Packet.yiaddr,GetOption(54));
                        StoreTime();
                        StoreOptions();
                        Handler(DHCP_REQUEST_SELECT,0,0);
                        return 1;
                };
                c++;
        };

        c = 0;  //can't get previous IP lets go on to get a new one
        while(c < n) {
                if(Request(0,ips[c],servers[c],1,1) > 0) {
                        StoreBasic(Packet.yiaddr,GetOption(54));
                        StoreTime();
                        StoreOptions();
                        Handler(DHCP_REQUEST_SELECT,0,0);
                        return 1;
                };
                c++;
        };*/

        nOffers = n;
        return n;
}

DHCP::DHCP()
{
}

int     DHCP::SetPrefTime(int t)
{
        if(t < 3600) PrefTime = 3600;
        else PrefTime = t;
        return PrefTime;
}

void    DHCP::Init(char *iface,int type,int len,const HWName &hwaddr,int append)
{
        DHCPLeaseData::Reset();
        DHCPSessionData::Reset();

        memset(BroadcastAddress,255,4);
        memset(NullcastAddress,0,4);
        ClientPort = DHCP_CLIENT_PORT;
        Options[0] = 255;
        ConformParms = 0;
        Socket = 0;
        PrefTime = 3600;
        DiscoverTimeout = 3000;
        RequestTimeout = 3000;

        int a = sizeof(Packet);
        if(a != 548) {
                fprintf(stderr,"Fatal error: Byte alignment problem in class 'DHCPMsg', out by %d\n",548 - sizeof(Packet));
                exit(1);
        };

        if(hwaddr != 0 && append == 0) {
                ConfigureInterface(iface,type,len,hwaddr.Address,hwaddr.AddressLength);
        }
        else {
                ConfigureInterface(iface,type,len,0,0);
        };

        if(append) {
                memcpy(&ClientID[CIDLength],hwaddr.Address,hwaddr.AddressLength);
                CIDLength += hwaddr.AddressLength;
        };

        Handler(DHCP_INIT,0,0);
}

DHCP::~DHCP()
{
        CloseChannel();
}

void    DHCP::SetConformity(int c)
{
        ConformParms = c;
}

void    DHCP::SetTimeouts(unsigned long d,unsigned long r)
{
        DiscoverTimeout = d;
        RequestTimeout = r;
}
