#include "includes.h"
#include "support.h"
#include "packet.h"

#ifdef __WIN32__
typedef int socklen_t;
#endif

#ifdef  DEBUG
#define DEBUG_CMD(s) s
#else
#define DEBUG_CMD(s)
#endif

#ifndef __WIN32__

int     EncodeIpEther(void *p,
                      int plen,
                      const unsigned char *dst,
                      const unsigned char *src)
{
        if(plen < (int) sizeof(ETHERHdr)) return 0;

        ETHERHdr &hdr = ((ETHERHdr *) p) [0];

        memcpy(hdr.DestAddr,dst,6);
        memcpy(hdr.SourceAddr,src,6);
        hdr.Type = htons(__ETHERTYPE_IP);

        return sizeof(hdr);
}

int     DecodeEther(void *p,int plen)
{
        if(plen < (int) sizeof(ETHERHdr)) return 0;
        return sizeof(ETHERHdr);
}

int     EncodeUdpIp(void *p,
                    int plen,
                    void *buf,
                    int blen,
                    unsigned short dstport,
                    const unsigned char *dstip,
                    unsigned short srcport,
                    const unsigned char *srcip)
{
        if(plen < (int) (blen + sizeof(IPHdr) + sizeof(UDPHdr))) return 0;

        struct  Packet {
                IPHdr    ih;
                UDPHdr   uh;
                char     Raw[1];
        };

        Packet &pkt = ((Packet *) p) [0];

        memcpy(pkt.Raw,buf,blen);
        memset(&pkt.ih,0,sizeof(IPHdr));

        memcpy(pkt.ih.SourceIP,srcip,4);
        memcpy(pkt.ih.DestIP,dstip,4);
        pkt.ih.Length = htons(blen + sizeof(UDPHdr));
        pkt.ih.Protocol = IPPROTO_UDP;

        pkt.uh.SourcePort = htons(srcport);
        pkt.uh.DestPort = htons(dstport);
        pkt.uh.Length = htons(blen + sizeof(UDPHdr));
        pkt.uh.Checksum = 0;
        pkt.uh.Checksum = CalcChecksum(&pkt.ih,blen + sizeof(IPHdr) + sizeof(UDPHdr),0);
        if(pkt.uh.Checksum == 0) pkt.uh.Checksum = 0xFFFF;

        pkt.ih.VerIHL = 0x45;
        pkt.ih.Length = htons(blen + sizeof(IPHdr) + sizeof(UDPHdr));
        pkt.ih.TTL = 64;
        pkt.ih.Checksum = 0;
        pkt.ih.Checksum = CalcChecksum(&pkt.ih,sizeof(IPHdr),0);

        return blen + sizeof(UDPHdr) + sizeof(IPHdr);
}

int     DecodeIp(void *p,int plen)
{
        if(plen < (int) sizeof(IPHdr)) return 0;
        IPHdr &ih = ((IPHdr *) p)[0];
        int ihl = (ih.VerIHL & 0xF) * 4;
        if(ihl > plen) return 0;
        if(ntohs(ih.Length) != plen) return 0;

        unsigned short csum = ih.Checksum;
        ih.Checksum = 0;
        ih.Checksum = CalcChecksum(p,ihl,0);
        if(ih.Checksum != csum) return 0;
        return ihl;
}

int     DecodeUdp(void *p,int plen,IPHdr &iphdr)
{
        if(iphdr.Protocol != IPPROTO_UDP) return 0;
        if(plen < (int) sizeof(UDPHdr)) return 0;
        UDPHdr &uh = ((UDPHdr *) p) [0];
        if(ntohs(uh.Length) != plen) return 0;

        if(uh.Checksum != 0) {
                IPHdr phdr;
                unsigned short csum = uh.Checksum;
                if(csum == 0xFFFF) csum = 0;
                uh.Checksum = 0;

                memset(&phdr,0,sizeof(phdr));
                memcpy(phdr.SourceIP,iphdr.SourceIP,4);
                memcpy(phdr.DestIP,iphdr.DestIP,4);
                phdr.Protocol = iphdr.Protocol;
                phdr.Length = uh.Length;

                uh.Checksum = CalcChecksum(&phdr,sizeof(phdr),p,plen,0);
                if(uh.Checksum != csum) {
                        return 0;
                };
        };

        return sizeof(uh);
}

#endif

int     IPHdr::Valid(int len)
{
        unsigned short csum = Checksum;
        Checksum = 0;
        if(csum == CalcChecksum(this,len,0)) {
                Checksum = csum;
                return 1;
        };
        Checksum = csum;
        return 0;
}

int     IPHdr::HdrLength()
{
        return (VerIHL & 0xF) * 4;
}

void *  IPHdr::End()
{
        return (void *) (((char *) this) + HdrLength());
}

IDF     ICMPPing::idvalue(1);

int     ICMPPing::Valid()
{
        unsigned short csum = Checksum;
        Checksum = 0;
        if(csum == CalcChecksum(this,16,0)) {
                Checksum = csum;
                return 1;
        };
        Checksum = csum;
        return 0;
}

int     ICMPPing::Ping(const unsigned char *ip,unsigned long wait,int persist)
{
        int s = socket(AF_INET,SOCK_RAW,IPPROTO_ICMP);
        unsigned char buf[512];
        socklen_t n;
        sockaddr sa;
        unsigned long l,t;
        fd_set fd;
        timeval tv;
        IPHdr &iphdr = ((IPHdr *) buf) [0];

        Type = 8;
        Code = 0;
        Checksum = 0;
        ID = idvalue.GetID();
        memcpy(GID,"IMPL1587",8);
        Checksum = CalcChecksum(this,16,0);

        int c = persist,d;
        while(c) {
                memset(&sa,0,sizeof(sa));
                sa.sa_family = AF_INET;
                memcpy(&sa.sa_data[2],ip,4);
                if(sendto(s,(char *) this,16,0,&sa,sizeof(sa)) <= 0) break;

                l = GetTicks();
                while((t = (GetTicks() - l)) < wait) {
                        t = wait - t;
                        tv.tv_sec = t / 1000UL;
                        tv.tv_usec = (t % 1000UL) * 1000UL;
                        FD_ZERO(&fd);
                        FD_SET(s,&fd);
                        if(select(s+1,&fd,0,0,&tv) == 0) break;
                        if(FD_ISSET(s,&fd) == 0) break;

                        n = sizeof(sa);
                        d = recvfrom(s,buf,512,0,&sa,&n);
                        ICMPPing &recvping = ((ICMPPing *) iphdr.End())[0];

                        if(ntohs(iphdr.Length) == d &&
                           (d - iphdr.HdrLength()) == 16 &&
                           iphdr.Valid(d) != 0 &&
                           recvping.Valid() != 0 &&
                           recvping.Type == 0 &&
                           recvping.ID == ID &&
                           memcmp(&sa.sa_data[2],ip,4) == 0 &&
                           memcmp(recvping.GID,"IMPL1587",8) == 0) {

#ifdef  __WIN32__
                                closesocket(s);
#else
                                close(s);
#endif
                                return 1;
                        };
                };
                c--;
        };
#ifdef  __WIN32__
        closesocket(s);
#else
        close(s);
#endif

        return 0;
}

unsigned short  CalcChecksum(void *par,...)
{
        struct arg {
                char *p;
                int n;
        };

        union conv {
                unsigned short s;
                unsigned char  b[2];
        };

        arg *p = ((arg *) &par);
        unsigned long   sum = 0;
        int touch = 0;

        while(p[0].p) {
                unsigned short *s = (unsigned short *) &p[0].p[touch];
                int len = p[0].n - touch;
                int c = 1;
                while(c < len) {
                        sum += s[0];
                        s++;
                        c++;c++;
                };
                if(c == len) {
                        if(p[1].p) {
                                conv b;
                                b.b[0] = ((unsigned char *) s) [0];
                                b.b[1] = ((unsigned char *) p[0].p) [0];
                                sum += b.s;
                        }
                        else sum += ((unsigned char *) s) [0];
                        touch = 1;
                }
                else touch = 0;

                p++;
        };
        sum = (sum >> 16) + (sum & 0xFFFF);
        sum += (sum >> 16);
        return ~sum;
}

#ifndef  __WIN32__

void            ARP::Out()
{
        puts("");
        printf("DHW:            %s\n",FmtHW(ETHERHdr::DestAddr,6));
        printf("SHW:            %s\n",FmtHW(ETHERHdr::SourceAddr,6));
        printf("hType:          %d\n",ntohs(hType));
        printf("pType:          %d\n",ntohs(pType));
        printf("hLength:        %d\n",hLength);
        printf("pLength:        %d\n",pLength);
        printf("op:             %d\n",ntohs(op));
        printf("sIP:            %s\n",ipname(sIP));
        printf("sHW:            %s\n",FmtHW(sHWAddress,6));
        printf("tIP:            %s\n",ipname(tIP));
        printf("tHW:            %s\n",FmtHW(tHWAddress,6));
        puts("");
}

unsigned char * ARP::GetHW(int s,
                           char *iface,
                           const unsigned char *destip,
                           const unsigned char *srcip,
                           const unsigned char *srchw,
                           int persistence)
{
        #define ARP_TIMEOUT     1000

        sockaddr        sa;             //receive/send address
        socklen_t       n;              //to receive length of sa
        int             c;              //count down var
        unsigned long   l,t;            //timing vars
        timeval         tv;             //wait interval temporary
        fd_set          fd;             //fd
        ARP             recvarp;

        memset(this,0,60);
        memcpy(ETHERHdr::DestAddr,"\xFF\xFF\xFF\xFF\xFF\xFF",6);
        memcpy(ETHERHdr::SourceAddr,srchw,6);
        ETHERHdr::Type          = htons(__ETHERTYPE_ARP);

        hType                   = htons(__ARPTYPE_ETHER);
        pType                   = htons(__ETHERTYPE_IP);
        hLength                 = 6;
        pLength                 = 4;
        op                      = htons(__ARP_REQUEST);
        memcpy(sHWAddress,srchw,6);
        memcpy(sIP,srcip,4);
        memcpy(tHWAddress,"\x0\x0\x0\x0\x0\x0",6);
        memcpy(tIP,destip,4);

        c = persistence;
        while(c) {
                memset(&sa,0,sizeof(sa));
                strcpy(sa.sa_data,iface);
                if(sendto(s,this,60,0,&sa,sizeof(sa)) <= 0) return 0;
                DEBUG_CMD(puts("\nSEND");)
                DEBUG_CMD(Out();)

                l = GetTicks();
                while((t = (GetTicks() - l)) < ARP_TIMEOUT) {
                        t = ARP_TIMEOUT - t;
                        tv.tv_sec = t / 1000UL;
                        tv.tv_usec = (t % 1000UL) * 1000UL;
                        FD_ZERO(&fd);
                        FD_SET(s,&fd);
                        if(select(s+1,&fd,0,0,&tv) == 0) break;
                        if(FD_ISSET(s,&fd) == 0) break;

/*                        n = sizeof(sa);
                        if(recvfrom(s,&recvarp,60,0,&sa,&n) > 0 &&
                           strcmp(sa.sa_data,iface) == 0 &&
                           htons(__ETHERTYPE_ARP) == recvarp.ETHERHdr::Type) {
                                recvarp.Out();
                        };*/

                        n = sizeof(sa);
                        if(recvfrom(s,&recvarp,60,0,&sa,&n) >= 42 &&
                           htons(__ETHERTYPE_ARP) == recvarp.ETHERHdr::Type &&
                           ntohs(recvarp.op) == __ARP_REPLY &&
                           strcmp(sa.sa_data,iface) == 0 &&
                           memcmp(recvarp.sIP,destip,4) == 0) {
                                DEBUG_CMD(recvarp.Out();)

                                memcpy(tHWAddress,recvarp.sHWAddress,6);
                                return tHWAddress;
                        };
                };

                c--;
        };

        return 0;
}

void    SendUnarp(int s,char *iface,unsigned char *hw,unsigned char *ip)
{
        struct Packet {
                ETHERHdr eh;
                unsigned short  hType;
                unsigned short  pType;
                unsigned char   hLength;
                unsigned char   pLength;
                unsigned short  op;
                unsigned char   sip[4];
                unsigned char   tip[4];
                unsigned char   PAD[30];
        };

        Packet p;
        sockaddr sa;

        memset(&sa,0,sizeof(sa));
        strcpy(sa.sa_data,iface);

        memset(p.eh.DestAddr,0xFF,6);
        memcpy(p.eh.SourceAddr,hw,6);
        p.eh.Type = htons(__ETHERTYPE_ARP);
        p.hType = htons(__ARPTYPE_ETHER);
        p.pType = htons(__ETHERTYPE_IP);
        p.hLength = 0;
        p.pLength = 4;
        p.op = htons(2);
        memcpy(p.sip,ip,4);
        memset(p.tip,0xFF,4);

        sendto(s,&p,sizeof(p),0,&sa,sizeof(sa));
}

#endif

