#ifndef __DHCPCLIENT__PACKET_H
#define __DHCPCLIENT__PACKET_H

#define __ARP_REQUEST           1
#define __ARP_REPLY             2
#define __ETHERTYPE_IP          0x0800
#define __ETHERTYPE_ARP         0x0806
#define __ARPTYPE_ETHER         1

unsigned short  CalcChecksum(void *,...);

class   ETHERHdr {
public:
        unsigned char   DestAddr[6];    //Destination address
        unsigned char   SourceAddr[6];  //Source address
        unsigned short  Type;           //NO, Type of packet
};

class   IPHdr {
public:
        //set ether type to htons(0x0800)

        unsigned char   VerIHL;         //0x45
        unsigned char   Tos;            //0
        unsigned short  Length;         //NO, Length of all headers and data (Network order)
        unsigned short  ID;             //NO, 0
        unsigned short  Offset;         //NO, 0 (Network order)
        unsigned char   TTL;            //254, Time to live
        unsigned char   Protocol;       //IPPROTO_UDP or IPPROTO_TCP
        unsigned short  Checksum;       //Checksum (Network order)
        unsigned char   SourceIP[4];    //Source address
        unsigned char   DestIP[4];      //Destination address

        int     Valid(int);
        int     HdrLength();
        void *  End();
};

class UDPHdr {
public:
        unsigned short  SourcePort;     //NO, Source port
        unsigned short  DestPort;       //NO, Dest port
        unsigned short  Length;         //NO, Header and data length
        unsigned short  Checksum;       //Checksum
};

class   ARP : public ETHERHdr {
private:
        unsigned short  hType;          //Hardware type
        unsigned short  pType;          //Protocol type
        unsigned char   hLength;        //Hardware address length
        unsigned char   pLength;        //Protocol address length
        unsigned short  op;             //Operator
        unsigned char   sHWAddress[6];  //Sendors Hardware address
        unsigned char   sIP[4];         //Sendors IP address
        unsigned char   tHWAddress[6];  //Target Hardware address
        unsigned char   tIP[4];         //Target IP address
        unsigned char   PAD[18];

public:
        void    Out();

        unsigned char *GetHW(int s,     //Socket to use
              char *,                   //iface binding
              const unsigned char *,          //Destination IP
              const unsigned char *,          //Source IP
              const unsigned char *,          //Source HW
              int);                     //Tries
};

class   ICMPPing {
private:
        unsigned char   Type;
        unsigned char   Code;
        unsigned short  Checksum;
        unsigned long   ID;
        unsigned char   GID[8];

        static   IDF    idvalue;

        int     Valid();
public:
        int     Ping(const unsigned char *,unsigned long,int);
};

#ifndef __WIN32__

int     EncodeIpEther(void *p,
                      int plen,
                      const unsigned char *dst,
                      const unsigned char *src);

int     DecodeEther(void *p,int plen);

int     EncodeUdpIp(void *p,
                    int plen,
                    void *buf,
                    int blen,
                    unsigned short dstport,
                    const unsigned char *dstip,
                    unsigned short srcport,
                    const unsigned char *srcip);

int     DecodeIp(void *p,int plen);

int     DecodeUdp(void *p,int plen,IPHdr &);

void    SendUnarp(int,char *,unsigned char *,unsigned char *);

#endif

#endif
